"""
Exceptions raised by cli-toolkit
"""


class ConfigurationError(Exception):
    """
    Errors raised by configuration processing
    """


class ScriptError(Exception):
    """
    Errors raise during script processing
    """



"""
Shared fixtures for unit tests
"""

import sys

import pytest


@pytest.fixture
def cli_mock_argv(monkeypatch):
    """
    Fixture to set argv for argparse tests
    """
    arguments = ['test-cli']
    monkeypatch.setattr(sys, 'argv', arguments)



"""
Loaders for ini, json and yaml format configuration files
"""

import configparser
import json
import os
import re

from pathlib import Path

import yaml

from sys_toolkit.logger import DEFAULT_TARGET_NAME

from .base import Base
from .exceptions import ConfigurationError

# Pattern to validate configuration keys
RE_CONFIGURATIION_KEY = re.compile('^[a-zA-Z0-9_]+$')


class ConfigurationItemContainer(Base):
    """
    Base class for containers of settings
    """
    __dict_loader_class__ = None
    __list_loader_class__ = None
    __float_settings__ = ()
    """Tuple of settings loaded as floats"""
    __integer_settings__ = ()
    """Tuple of settings loaded as integers"""
    __path_settings__ = ()
    """Tuple of settings loaded as pathlib.Path"""

    def __init__(self, parent=None, debug_enabled=False, silent=False, logger=DEFAULT_TARGET_NAME):
        self.__attributes__ = []
        super().__init__(parent, debug_enabled, silent, logger)

    @property
    def __config_root__(self):
        """
        Return configuration root item via parent links
        """
        parent = getattr(self, '__parent__', None)
        if parent is None:
            return self
        while getattr(parent, '__parent__', None) is not None:
            parent = getattr(parent, '__parent__', None)
        return parent

    @property
    def __dict_loader__(self):
        """
        Return loader for dict items
        """
        if self.__dict_loader_class__ is not None:
            return self.__dict_loader_class__
        return ConfigurationSection

    @property
    def __list_loader__(self):
        """
        Return loader for list items
        """
        if self.__list_loader_class__ is not None:
            return self.__list_loader_class__
        return ConfigurationList

    @staticmethod
    def default_formatter(value):
        """
        Default formatter for variables

        By default trims whitespace from strings and sets empty strings to None
        """
        if isinstance(value, str):
            value = value.strip()
            if value == '':
                value = None
        return value

    @staticmethod
    def __validate_attribute__(attr):
        """
        Validate attribute to be set
        """
        if not isinstance(attr, str):
            raise ConfigurationError(f'Attribute is not string: {attr}')
        if not attr.isidentifier():
            raise ConfigurationError(f'Attribute is not valid python identifier: {attr}')
        if not RE_CONFIGURATIION_KEY.match(attr):
            raise ConfigurationError(f'Invalid attribute name: {attr}')

    def as_dict(self):
        """
        Return VS code configuration section as dictionary
        """
        data = {}
        for attribute in self.__attributes__:
            item = getattr(self, attribute)
            if hasattr(item, 'as_dict'):
                data[attribute] = item.as_dict()
            else:
                data[attribute] = item
        return data

    def set(self, attr, value):
        """
        Load item with correct class
        """
        self.__validate_attribute__(attr)

        section = getattr(self, attr, None)
        if section is not None and callable(getattr(section, 'set', None)):
            section.set(attr, value)
            return

        if isinstance(value, dict):
            item = self.__dict_loader__(value, parent=self)  # pylint: disable=not-callable
            setattr(self, attr, item)
            self.__attributes__.append(attr)
            return

        if isinstance(value, (list, tuple)):
            item = self.__list_loader__(attr, value, parent=self)  # pylint: disable=not-callable
            setattr(self, item.__setting__, item)
            self.__attributes__.append(item.__setting__)
            return

        if value is not None:
            if attr in self.__float_settings__:
                value = float(value)
            if attr in self.__integer_settings__:
                value = int(value)
            if attr in self.__path_settings__:
                value = Path(value).expanduser()

        validator_callback = getattr(self, f'validate_{attr}', None)
        if callable(validator_callback):
            try:
                value = validator_callback(value)  # pylint: disable=not-callable
            except Exception as error:
                raise ConfigurationError(f'Error validating setting {attr}: {error}') from error

        formatter_callback = getattr(self, f'format_{attr}', None)
        try:
            if callable(formatter_callback):
                value = formatter_callback(value)  # pylint: disable=not-callable
            else:
                value = self.default_formatter(value)
        except Exception as error:
            raise ConfigurationError(f'Error formatting setting {attr}: {error}') from error

        setattr(self, attr, value)
        self.__attributes__.append(attr)


class ConfigurationList(ConfigurationItemContainer):
    """
    List of settings in configuration section
    """
    def __init__(self, setting=None, data=None, parent=None):
        super().__init__(parent=parent)
        self.__setting__ = setting
        self.__load__(data)

    def __repr__(self):
        return self.__values__.__repr__()

    def __getitem__(self, index):
        return self.__values__[index]

    def __iter__(self):
        return iter(self.__values__)

    def __len__(self):
        return len(self.__values__)

    def __load__(self, value):
        """
        Load list of values
        """
        self.__values__ = []
        if value:
            for item in value:
                if isinstance(item, dict):
                    # pylint: disable=not-callable
                    item = self.__dict_loader__(item, parent=self)
                self.__values__.append(item)

    def set(self, attr, value):
        self.__load__(value)


class ConfigurationSection(ConfigurationItemContainer):
    """
    Configuration section with validation

    Configuration sections can be nested and linked to parent configuration
    section by parent argument. If parent is given it must be an instance of
    ConfigurationSection.

    Any data given in data dictionary are inserted as settings.
    """
    __name__ = None
    """Name of configuration section, used in linking custom classes"""
    __default_settings__ = {}
    """Dictionary of default settings configuration"""
    __required_settings__ = ()
    """Tuple of settings required for valid configuration"""
    __environment_variables__ = {}
    """Mapping from environment variables read to settings"""
    __environment_variable_prefix__ = None
    """Prefix for reading settings from environment variables"""

    __section_loaders__ = ()
    """Classes used for subsection parsers"""
    __key_attribute_map__ = {}
    """Map configuration keys to python compatible attributes"""

    def __init__(self, data=dict, parent=None, debug_enabled=False, silent=False):
        if parent is not None and not isinstance(parent, ConfigurationItemContainer):
            raise TypeError('parent must be instance of ConfigurationItemContainer')
        super().__init__(
            parent=parent,
            debug_enabled=debug_enabled,
            silent=silent,
        )

        self.__subsections__ = []

        self.__valid_settings__ = self.__detect_valid_settings__()
        for attr in self.__valid_settings__:
            self.set(attr, None)

        self.__initialize_sub_sections__()
        self.__load_dictionary__(self.__default_settings__)
        if isinstance(data, dict):
            self.__load_dictionary__(data)

        self.__load_environment_variables__()

        if parent is None:
            self.validate()

    def __repr__(self):
        return self.__name__ if self.__name__ is not None else ''

    def __initialize_sub_sections__(self):
        """
        Initialize sub sections configured in __section_loaders__
        """
        for loader in self.__section_loaders__:
            subsection = loader(data=None, parent=self)
            name = self.__attribute_from_key__(subsection.__name__)
            if name is None:
                raise ConfigurationError(f'Subsection class defines no name: {loader}')
            setattr(self, name, subsection)
            self.__attributes__.append(name)

    def __attribute_from_key__(self, key):
        """
        Map settings key to python attribute
        """
        return self.__key_attribute_map__.get(key, key)

    def __key_from_attribute__(self, attr):
        """
        Map settings file key from attribute
        """
        for key, value in self.__key_attribute_map__.items():
            if attr == value:
                return key
        return attr

    def __split_attribute_path__(self, key):
        """
        Return section attribute from key
        """
        key = self.__attribute_from_key__(key)
        if isinstance(key, str):
            parts = key.split('.')
            return parts[0], '.'.join(parts[1:])
        return key, []

    def __detect_valid_settings__(self):
        """
        Detect all known settings, return list of keys
        """
        attributes = []
        attributes = list(self.__required_settings__)
        for attr, value in self.__default_settings__.items():
            if not RE_CONFIGURATIION_KEY.match(attr):
                raise ConfigurationError(f'Invalid attribute name: {attr}')
            if not isinstance(value, dict) and attr not in attributes:
                attributes.append(attr)
        for attr in self.__environment_variables__.values():
            if not RE_CONFIGURATIION_KEY.match(attr):
                raise ConfigurationError(f'Invalid attribute name: {attr}')
            if attr not in attributes:
                attributes.append(attr)
        return sorted(set(attributes))

    def __load_environment_variables__(self):
        """
        Load settings from environment variables
        """
        if self.__environment_variable_prefix__ is not None:
            for attr in self.__valid_settings__:
                env = f'{self.__environment_variable_prefix__}_{attr}'.upper()
                value = os.environ.get(env, None)
                if value is not None:
                    self.set(attr, value)

        for env, attr in self.__environment_variables__.items():
            value = os.environ.get(env, None)
            if value is not None:
                self.set(attr, value)

    def __get_section_loader__(self, section_name):
        """
        Find configuration section loader by name

        By default returns ConfigurationSection
        """
        if not isinstance(section_name, str) or section_name == '':
            raise ConfigurationError('Configuration section name not defined')

        section_name = self.__attribute_from_key__(section_name)
        for loader in self.__section_loaders__:
            loader_name = self.__attribute_from_key__(getattr(loader, '__name__', None))
            if loader_name == section_name:
                return loader
        return self.__dict_loader__

    def __get_or_create_subsection__(self, name, parent=None):
        if parent is None:
            parent = self
        if not hasattr(parent, name):
            loader = parent.__get_section_loader__(name)
            item = loader({}, parent=parent)
            item.__name__ = name
            setattr(parent, name, item)
            parent.__subsections__.append(item)
        return getattr(parent, name)

    def __init_subsection_path__(self, section_name, path):
        """
        Initialize subsections from config path

        Splits . separated path to config setting path, creates subsections
        on path excluding last component

        Returns subsection matching longest path and field (last path component).
        Field is not actually loaded as the dictionary key
        """
        subsection = self.__get_or_create_subsection__(section_name)
        path = path.split('.')
        field = path[-1]
        if len(path) > 1:
            for subsection_name in path[:-1]:
                subsection = self.__get_or_create_subsection__(
                    subsection_name,
                    parent=subsection
                )
        return subsection, field

    def __load_section__(self, section, data, path=None):
        """
        Load configuration section from data
        """
        if path is not None:
            subsection, item = self.__init_subsection_path__(section, path)
            if isinstance(data, dict):
                if item == path:
                    subsection = self.__get_or_create_subsection__(
                        item,
                        parent=subsection
                    )
                subsection.__load_dictionary__(data)
            else:
                subsection.set(item, data)
        elif isinstance(data, dict):
            subsection = self.__get_or_create_subsection__(section, parent=self)
            subsection.__load_dictionary__(data)
        else:
            raise ConfigurationError('not a dict')

    def __load_dictionary__(self, data):
        """
        Load settings from dictionary

        Any dictionaries in data are loaded as child configuration sections
        """
        if not isinstance(data, dict):
            raise ConfigurationError(f'Dictionary is not dict instance: {data}')

        for key, value in data.items():
            attr, path = self.__split_attribute_path__(key)
            if path:
                self.__load_section__(attr, value, path)
            elif isinstance(value, dict):
                self.__load_section__(attr, value)
            else:
                self.set(key, value)

    def as_dict(self):
        """
        Return configuration section as dictionary
        """
        data = super().as_dict()
        for subsection in self.__subsections__:
            data[subsection.__name__] = subsection.as_dict()
        return data

    def set(self, attr, value):
        """
        Set configuration setting value as attribute of the object

        Attributes are set as ConfigurationItem classes by default.
        """
        attr, path = self.__split_attribute_path__(attr)
        if path:
            self.__load_section__(attr, value, path)
            return

        self.__validate_attribute__(attr)
        super().set(attr, value)

    def validate(self):
        """
        Validate loaded configuration settings

        Default implementation checks if required settings are set.
        """
        for attr in self.__required_settings__:
            value = getattr(self, attr, None)
            if value is None:
                raise ConfigurationError(f'{self} required setting {attr} has no value')


class ConfigurationFile(ConfigurationSection):
    """
    Common base class for configuration file parsers
    """
    __default_paths__ = []

    def __init__(self, path=None, parent=None, debug_enabled=False, silent=False):
        self.__path__ = Path(path).expanduser() if path is not None else None
        super().__init__(parent=parent, debug_enabled=debug_enabled, silent=silent)

        # Load any common default paths
        for default_path in self.__default_paths__:
            default_path = Path(default_path)
            if default_path.is_file():
                self.load(default_path)

        # Load specified configuration file
        if self.__path__ is not None and self.__path__.exists():
            self.load(self.__path__)

    def __repr__(self):
        return str(self.__path__.name) if self.__path__ is not None else ''

    @staticmethod
    def __check_file_access__(path):
        """
        Check access to specified path as file

        Raises ConfigurationError if path is not a file or not readable
        """
        path = Path(path).expanduser()
        if not path.is_file():
            raise ConfigurationError(f'No such file: {path}')
        if not os.access(path, os.R_OK):
            raise ConfigurationError(f'Permission denied: {path}')
        return path

    def load(self, path):
        """
        Load specified configuration file

        This method must be implemted in child class
        """
        raise NotImplementedError('File loading must be implemented in child class')

    def parse_data(self, data):
        """
        Parse data read from configuration file

        Default implementation requires data is a dictionary
        """
        if data is None:
            return
        if not isinstance(data, dict):
            raise ConfigurationError(f'Data is not dict instance: {data}')
        self.__load_dictionary__(data)


class IniConfiguration(ConfigurationFile):
    """
    Configuration parser for INI style configuration files

    You can pass arguments to configparser.ConfigParser with
    loader_args
    """
    def __init__(self, path=None, parent=None, debug_enabled=False, silent=False, **loader_args):
        self.__loader_args__ = loader_args
        super().__init__(path, parent=parent, debug_enabled=debug_enabled, silent=silent)

    def load(self, path):
        """
        Load specified INI configuration file
        """
        path = self.__check_file_access__(path)
        parser = configparser.ConfigParser(**self.__loader_args__)
        try:
            parser.read(path)
        except Exception as error:
            raise ConfigurationError(f'Error loading {path}: {error}') from error

        for section_name in parser.sections():
            self.__load_section__(
                section_name,
                dict(parser[section_name].items())
            )


class JsonConfiguration(ConfigurationFile):
    """
    Configuration parser for JSON configuration files

    You can pass arguments to json.loads with loader_args
    """
    encoding = 'utf-8'

    def __init__(self, path=None, parent=None, debug_enabled=False, silent=False, **loader_args):
        self.__loader_args__ = loader_args
        super().__init__(path, parent=parent, debug_enabled=debug_enabled, silent=silent)

    def load(self, path):
        """
        Load specified JSON configuration file
        """
        path = self.__check_file_access__(path)
        try:
            self.parse_data(
                json.loads(path.open('r', encoding=self.encoding).read(), **self.__loader_args__)
            )
        except Exception as error:
            raise ConfigurationError(f'Error loading {path}: {error}') from error


class YamlConfiguration(ConfigurationFile):
    """
    Configuration parser for yaml configuration files
    """
    encoding = 'utf-8'

    def load(self, path):
        """
        Load specified YAML configuration file
        """
        path = self.__check_file_access__(path)

        try:
            self.parse_data(
                yaml.safe_load(path.open('r', encoding='utf-8'))
            )
        except Exception as error:
            raise ConfigurationError(f'Error loading {path}: {error}') from error



"""
CLI script base class

This module implements the common base class to quickly create CLI commands
with python.

The Script class is usually used with cli_toolkit.command.Command to create
CLI scripts with subcommands.
"""

import argparse
import signal
import sys

from pathlib import Path

from setproctitle import setproctitle  # pylint: disable=no-name-in-module

from sys_toolkit.logger import Logger

from .base import NestedCliCommand


class ScriptMetaClass(type):
    """
    Run script initialize() method after creating object
    """
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.initialize()
        return obj


class Script(NestedCliCommand, metaclass=ScriptMetaClass):
    """
    CLI script command main class
    """
    subcommands = ()

    def __init__(self, usage=None, description=None, epilog=None, formatter_class=None):
        signal.signal(signal.SIGINT, self.SIGINT)
        self.name = Path(sys.argv[0]).name
        self.logger = Logger(self.name)
        super().__init__()

        setproctitle(f'{self.name} {" ".join(sys.argv[1:])}')

        if formatter_class is None:
            formatter_class = self.default_formatter_class

        self.__parser__ = argparse.ArgumentParser(
            prog=self.name,
            usage=usage,
            description=description,
            epilog=epilog,
            formatter_class=formatter_class,
            add_help=True,
            conflict_handler='resolve',
        )

        self.__parser__.add_argument('--debug', action='store_true', help='Enable debug messages')
        self.__parser__.add_argument('--quiet', action='store_true', help='Silent printed messages')

    # pylint: disable=unused-argument
    def initialize(self, *args, **kwargs):
        """
        Add subcommands defined in self.subcommands after creating object instance
        """
        self.register_parser_arguments(self.__parser__)
        self.__register_subcommands__()

    # pylint: disable=invalid-name
    # pylint: disable=unused-argument
    def SIGINT(self, signum, frame):
        """
        Parse SIGINT signal by quitting the program cleanly with exit code 1
        """
        self.reset_stty()
        self.exit(1)

    def __process_args__(self, args):
        """
        Process args and run subcommand if detected
        """
        if getattr(args, 'debug', None):
            self.__debug_enabled__ = True

        if getattr(args, 'quiet', None):
            self.__silent__ = True

        return args

    def parse_args(self):
        """
        Call parse_args for parser and check for default logging flags
        """
        return self.__process_args__(self.__parser__.parse_args())

    def parse_known_args(self):
        """
        Call parse_args for parser and check for default logging flags
        """
        args, other_args = self.__parser__.parse_known_args()
        args = self.__process_args__(args)
        return args, other_args

    def run(self):
        """
        Run script, parsing arguments and running subcommands

        This expects subcommands have been registered and exists if not

        This simply runs self.parse_args() not expecting to do anything with
        returned values since the subcommand is run.
        """
        args = self.parse_args()
        self.run_subcommand(args)



"""
Common unit test validators for cli_toolkit.script.Script objects
"""

import argparse
import sys

import pytest

from ..script import Script


def validate_script_attributes(script,
                               expected_args,
                               formatter_class=None,
                               debug_enabled=False
                               ):
    """
    Validate common attributes of Script instance
    """
    assert isinstance(script, Script)

    assert script.__debug_enabled__ == debug_enabled

    if formatter_class is not None:
        assert script.__parser__.formatter_class == formatter_class
    else:
        assert script.__parser__.formatter_class == script.default_formatter_class

    args = script.parse_args()
    assert isinstance(args, argparse.Namespace)
    assert vars(args) == expected_args


def validate_script_error_exit_code(exception, exit_code=1):
    """
    Assert running script raises specified exit code
    """
    assert exception.type == SystemExit
    assert exception.value.code == exit_code
    return exception


def validate_script_run_exception_with_args(script, context, args, exit_code=2):
    """
    Validate exception raised by running script with specified invalid
    arguments
    """
    context.setattr(sys, 'argv', args)
    with pytest.raises(SystemExit) as exception:
        script.run()
    assert exception.value.code == exit_code


def validate_script_debug_flag_enabled(script, context, expected_args):
    """
    Validate debug flag is set by script correctly
    """
    script.add_argument('--test', action='store_true', help='Set test flag')
    expected_args['debug'] = True
    expected_args['test'] = True
    testargs = [sys.argv[0], '--debug', '--test']
    context.setattr(sys, 'argv', testargs)
    args = script.parse_args()
    assert vars(args) == expected_args


def validate_script_quiet_flag_enabled(script, context, expected_args):
    """
    Validate debug flag is set by script correctly
    """
    script.add_argument('--test', action='store_true', help='Set test flag')
    expected_args['quiet'] = True
    expected_args['test'] = True
    testargs = [sys.argv[0], '--quiet', '--test']
    context.setattr(sys, 'argv', testargs)
    args = script.parse_args()
    assert vars(args) == expected_args



"""
CLI utility toolkit

This module implements python modules to implement command line tools,
extending python argparse.ArgumentParser to make writing nested CLI
commands easier.
"""
__version__ = '1.2.0'



"""
Asynchronous tasks for CLI scripts

This module implements script tasks to be executed asynchronously and
concurrently from script's run() method.
"""

import asyncio
import locale


class BaseScriptTask:
    """
    Common base class for script tasks

    :param parent: Parent script or command
    :type parent: Script, Command

    :param kwargs: Arguments passed to the task when run
    :type kwargs: dict
    """
    def __init__(self, parent, **kwargs):
        self.parent = parent
        self.messages = []
        self.errors = []
        parent.add_async_task(self.run, **kwargs)

    def error(self, *args):
        """
        Send subtask errors to parent
        """
        return self.parent.error(*args)

    def message(self, *args):
        """
        Send subtask messages to arent
        """
        return self.parent.message(*args)

    async def run(self, **kwargs):
        """
        Run task. This must be implemented in child class
        """
        raise NotImplementedError('Task run() must be impelemented in child class')


class CommandLineTask(BaseScriptTask):
    """
    Shell command task linked to CLI script or command

    Runs a non-interactive shell command with specified arguments
    """
    def __init__(self, parent, command, **kwargs):
        super().__init__(parent, **kwargs)
        self.command = command

    async def process_stderr(self, stderr):
        """
        Method to process asynchronous messages to stderr from process
        """
        async for line in stderr:
            error = line.decode(locale.getpreferredencoding(False)).rstrip()
            self.errors.append(error)
            self.error(error)

    async def process_stdout(self, stdout):
        """
        Method to process asynchronous messages to stdout from process
        """
        async for line in stdout:
            message = line.decode(locale.getpreferredencoding(False)).rstrip()
            self.messages.append(message)
            self.message(message)

    async def run(self, **kwargs):
        """
        Run specified shell command with asyncio
        """
        process = await asyncio.create_subprocess_exec(
            *self.command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await self.process_stdout(process.stdout)
        await self.process_stderr(process.stderr)
        return await process.wait()


class Task(BaseScriptTask):
    """
    Script python task linked to CLI script or command

    This class must be inherited to child class that implements run() method.
    """

    async def run(self, **kwargs):
        """
        Run task. This must be implemented in child class
        """
        raise NotImplementedError('Task run() must be impelemented in child class')



"""
Base classes for scripts and script subcommands
"""

import argparse
import asyncio
import os
import sys

from sys_toolkit.base import LoggingBaseClass

from .exceptions import ScriptError

DEFAULT_SUBPARSER_HELP = ''


class Base(LoggingBaseClass):
    """
    Base class with message / error handling and runner for registered
    async tasks

    Also includes self.logger instance of cli_toolkit.logger.Logger with
    specified logger group name
    """
    def __init__(self, parent=None, debug_enabled=False, silent=False, logger=None):
        super().__init__(debug_enabled, silent, logger)
        if parent is not None and not isinstance(parent, Base):
            raise TypeError('parent must be instance of cli_toolkit.base.Base')
        self.__parent__ = parent
        self.__async_task_callbacks__ = []
        self.__async_tasks__ = []

    @property
    def __is_debug_enabled__(self):
        """
        Check if debugging is enabled in this class or parents
        """
        if self.__parent__ is not None:
            return self.__parent__.__is_debug_enabled__ or self.__debug_enabled__
        return super().__is_debug_enabled__

    @property
    def __is_silent__(self):
        """
        Check if silent mode is enabled in this class or parents
        """
        if self.__parent__ is not None:
            return self.__parent__.__is_silent__ or self.__silent__
        return super().__is_silent__

    def add_async_task(self, callback, **kwargs):
        """
        Register a new async task to be run when run_async_tasks is triggered
        """
        self.__async_task_callbacks__.append((callback, kwargs))

    async def create_async_tasks(self):
        """
        Create asynchronous tasks added by self.add_async_task_callback
        """
        self.__async_tasks__ = []

        for callback_args in self.__async_task_callbacks__:
            callback = callback_args[0]
            kwargs = callback_args[1]
            task = asyncio.create_task(callback(**kwargs))
            self.__async_tasks__.append(task)

        for task in self.__async_tasks__:
            await task

        return asyncio.gather(*self.__async_tasks__)

    def run_async_tasks(self):
        """
        Create and run async tasks registered with add_async_task
        """
        return asyncio.run(self.create_async_tasks())


class NestedCliCommand(Base):
    """
    Common base class for scripts and nested CLI commands
    """
    name = None
    """Name of script or command, used in CLI command arguments generation"""
    usage = ''
    """Command usage string for --help in ArgumentParser"""
    description = ''
    """Command description, shown on top of --help output"""
    epilog = ''
    """Command epilog, shown on botton of --help output"""
    default_formatter_class = argparse.RawTextHelpFormatter
    """Argument parser formatter class"""
    no_subcommand_error = 'No command selected'
    """Error message to show when no required subcommand is specified when running CLI"""

    subcommands = ()

    def __init__(self, parent=None):
        """
        Initialize command with link t parent

        Parent can be either Script or Command
        """

        if parent is not None and not isinstance(parent, NestedCliCommand):
            raise TypeError('parent must be instance of NestedCliCommand')

        super().__init__(parent=parent)
        self.__subcommand_parser__ = None
        self.__subcommands__ = {}

        self.__parser__ = parent.__parser__ if parent is not None else None

        if self.name is None:
            raise ScriptError(f'No name defined: {self.__class__}')

    def __repr__(self):
        """
        Return name of script or command
        """
        return self.name if self.name is not None else ''

    @property
    def command_dest(self):
        """
        Return ArgumentParser 'dest' flag for command
        """
        return f'{self.name}_command'

    def exit(self, value=0, message=None):
        """
        Exit the script with given exit value

        If message is not None, it is output to stderr
        """
        if isinstance(value, bool):
            value = 0 if value else 1
        else:
            try:
                value = int(value)
                if value < 0 or value > 255:
                    raise ValueError
            except ValueError:
                value = 1

        if message:
            self.error(message)

        for task in self.__async_tasks__:
            task.cancel()

        self.reset_stty()
        sys.exit(value)

    def __register_subcommands__(self):
        """
        Register nested subcommands
        """
        for loader in self.subcommands:
            command = loader(self)
            if command.name not in self.__subcommands__:
                self.add_subcommand(command)
            command.__register_subcommands__()

    @staticmethod
    def reset_stty():
        """
        Clear stty settings by running 'stty sane' in terminal
        """
        if sys.platform == 'win32':
            return
        try:
            os.system('stty sane')
        except Exception:
            pass

    # pylint: disable=redefined-builtin
    def add_subparsers(self, help=None):
        """
        Add subparsers linked to parent parser
        """
        assert self.__parser__ is not None
        assert self.__subcommand_parser__ is None

        if help is None:
            help = DEFAULT_SUBPARSER_HELP

        self.__subcommand_parser__ = self.__parser__.add_subparsers(
            dest=self.command_dest,
            help=help,
        )
        self.__subcommands__ = {}

    # pylint: disable=redefined-builtin
    def add_subcommand(self, command, help=None, formatter_class=None):
        """
        Add a subcommand parser linked to nested command
        """
        if self.__subcommand_parser__ is None:
            self.add_subparsers(help=help)

        if formatter_class is None:
            formatter_class = self.default_formatter_class

        if command.name in self.__subcommands__:
            self.exit(1, f'Subcommand already registered: {self} {command.name}')

        parser = self.__subcommand_parser__.add_parser(
            name=command.name,
            usage=getattr(command, 'usage', None),
            description=getattr(command, 'description', None),
            epilog=getattr(command, 'epilog', None),
            formatter_class=formatter_class,
        )
        command.__parser__ = parser
        self.__subcommands__[command.name] = command

        parser = command.register_parser_arguments(parser)

        command.__register_subcommands__()
        return parser

    def add_argument(self, *args, **kwargs):
        """
        Shortcut to add argument to main argumentparser instance
        """
        self.__parser__.add_argument(*args, **kwargs)

    @staticmethod
    def register_parser_arguments(parser):
        """
        Register parser arguments

        By default does nothing
        """
        return parser

    def run_subcommand(self, args):
        """
        Run subcommand with arguments
        """
        if self.__async_task_callbacks__:
            self.run_async_tasks()
            self.exit(0)

        if self.command_dest not in args:
            self.exit(1, 'Command defines no subcommands')

        command_dest = getattr(args, self.command_dest, None)
        if command_dest is None:
            self.exit(1, self.no_subcommand_error)

        command = self.__subcommands__[command_dest]
        args = command.parse_args(args)
        command.run(args)
        # Explicitly exit after running command
        self.exit(0)



"""
CLI script subcommands

Subcommands can be linked to main Script instance or as nested commands to
Command instances.
"""

import argparse

from .base import NestedCliCommand


class Command(NestedCliCommand, argparse.ArgumentParser):
    """
    CLI subcommand

    Each CLI subcommand is initialized with parent reference. Parent must be instance
    of NestedCliCommand, i.e. Script or Command

    Arguments usage, description and epilog can be ignored and passed in child class
    as class attributes
    """

    def __init__(self, parent, usage=None, description=None, epilog=None):
        super().__init__(parent=parent)
        self.prog = self.name
        self.usage = self.__get_field__('usage', usage)
        self.description = self.__get_field__('description', description)
        self.epilog = self.__get_field__('epilog', epilog)

    def __get_field__(self, field, value=None):
        """
        Get value for field from specified value, class attribute or using default

        If value is not not empty the command tries to use class attribute value
        """
        if value:
            return value
        return getattr(self, field)

    def exit(self, value=0, message=None):
        """
        Pass exit() call to parent command or script
        """
        self.__parent__.exit(value, message)

    def debug(self, *args):
        """
        Pass debug() call to parent command or script
        """
        self.__parent__.debug(*args)

    def error(self, *args):
        """
        Pass error() call to parent command or script
        """
        self.__parent__.error(*args)

    def message(self, *args):
        """
        Pass message() call to parent command or script
        """
        self.__parent__.message(*args)

    def parse_args(self, args=None, namespace=None):
        """
        Parse command arguments

        Command arguments are received from parent and not parsed
        here with ArgumentParser default parser.
        """
        return args

    def parse_known_args(self, args=None, namespace=None):
        """
        Parse command arguments with unknown arguments

        Command arguments are received from parent and not parsed
        here with ArgumentParser default parser.
        """
        return args, namespace

    def run(self, args):
        """
        Run command with arguments from argument parser
        """
        return self.run_subcommand(args)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cli_toolkit', 'cli_toolkit.tests']

package_data = \
{'': ['*']}

install_requires = \
['setproctitle>=1.2.2,<2.0.0', 'sys-toolkit>=1.2.0,<2.0.0']

entry_points = \
{'pytest11': ['cli_toolkit_fixtures = cli_toolkit.fixtures']}

setup_kwargs = {
    'name': 'cli-toolkit',
    'version': '1.2.0',
    'description': 'Classes to implement CLI commands in python',
    'long_description': None,
    'author': 'Ilkka Tuohela',
    'author_email': 'hile@iki.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
